@@ -28,7 +28,7 @@ class RelationStateManager(StateBase):
         atomically.
         """
 
-        # the TOODs in the following comments in this function are for
+        # the TODOs in the following comments in this function are for
         # type checking to be implemented ASAP. However, this will
         # require some nontrivial test modification, so it's best to
         # be done in a future branch. This is because these tests use
@@ -58,25 +58,39 @@ class RelationStateManager(StateBase):
         if topology.has_relation_between_endpoints(endpoints):
             raise RelationAlreadyExists(endpoints)
 
+        scope = "global"
+        for endpoint in endpoints:
+            if endpoint.relation_scope == "container":
+                scope = "container"
+                break
+
         relation_type = endpoints[0].relation_type
-        relation_id = yield self._add_relation_state(relation_type)
+        relation_id = yield self._add_relation_state(relation_type, scope)
+
         services = []
         for endpoint in endpoints:
-            service_id = topology.find_service_with_name(endpoint.service_name)
+            service_id = topology.find_service_with_name(
+                endpoint.service_name)
+
             yield self._add_service_relation_state(
-                relation_id, service_id, endpoint)
+                relation_id, service_id, endpoint, scope)
+
             services.append(ServiceRelationState(self._client,
                                                  service_id,
                                                  relation_id,
+                                                 scope,
                                                  endpoint.relation_role,
                                                  endpoint.relation_name))
 
         def add_relation(topology):
             if topology.has_relation_between_endpoints(endpoints):
                 raise RelationAlreadyExists(endpoints)
-            topology.add_relation(relation_id, relation_type)
+            topology.add_relation(relation_id, relation_type,
+                                  relation_scope=scope)
             for service_relation in services:
-                if not topology.has_service(service_id):
+                if not topology.has_service(
+                    service_relation.internal_service_id):
+                    # One of the service endpoints has gone away.
                     raise StateChanged()
                 topology.assign_service_to_relation(
                     relation_id,
@@ -88,30 +102,36 @@ class RelationStateManager(StateBase):
         returnValue((RelationState(self._client, relation_id), services))
 
     @inlineCallbacks
-    def _add_relation_state(self, relation_type):
+    def _add_relation_state(self, relation_type, relation_scope):
         path = yield self._client.create(
             "/relations/relation-", flags=zookeeper.SEQUENCE)
         internal_id = basename(path)
-        # create the settings container, for individual units settings.
-        yield self._client.create(path + "/settings")
+
+        # Create the settings container, for individual units settings.
+        # Creation is per container for container scoped relations and
+        # occurs elsewhere.
+        if relation_scope == "global":
+            yield self._client.create(path + "/settings")
         returnValue(internal_id)
 
     @inlineCallbacks
     def _add_service_relation_state(
-        self, relation_id, service_id, endpoint):
+        self, relation_id, service_id, endpoint, relation_scope):
         """Add a service relation state.
         """
-        # Add service container in relation.
-        node_data = yaml.safe_dump(
-            {"name": endpoint.relation_name, "role": endpoint.relation_role})
-        path = "/relations/%s/%s" % (relation_id, endpoint.relation_role)
-
+        # while the full path is
+        # /relations/relation_id/optional_container_id/relation_role/...
+        # how far down the path we can create at this point depends on
+        # what type of relation we have.
+        if relation_scope == "global":
+            # Add service container in relation.
+            path = "/relations/%s/%s" % (relation_id, endpoint.relation_role)
+        else:
+            path = "/relations/%s" % (relation_id,)
         try:
-            yield self._client.create(path, node_data)
+            yield self._client.create(path)
         except zookeeper.NodeExistsException:
-            # If its not, then update the node, and continue.
-            yield retry_change(
-                self._client, path, lambda content, stat: node_data)
+            pass
 
     @inlineCallbacks
     def remove_relation_state(self, relation_state):
@@ -144,12 +164,13 @@ class RelationStateManager(StateBase):
         internal_service_id = service_state.internal_id
         topology = yield self._read_topology()
         for info in topology.get_relations_for_service(internal_service_id):
-            relation_id, relation_type, service_info = info
+            service_info = info["service"]
             relations.append(
                 ServiceRelationState(
                     self._client,
                     internal_service_id,
-                    relation_id,
+                    info["relation_id"],
+                    info["scope"],
                     **service_info))
         returnValue(relations)
 
@@ -182,24 +203,60 @@ class RelationState(StateBase):
         return self._internal_id
 
 
-class ServiceRelationState(object):
+class ServiceRelationState(StateBase):
     """A state representative of a relation between one or more services."""
 
-    def __init__(self, client, service_id, relation_id, role, name):
-        self._client = client
+    def __init__(self, client, service_id, relation_id, relation_scope,
+                 role, name):
+        """`client`: ZooKeeper client
+        `service_id`: service internal id
+        `relation_id`: relation internal id
+        `relation_scope`: the active scope of the service relation (global or
+        container)
+        `role`: the role of of the service in the relation
+        `name`: the name of the service in the relation
+        """
+        super(ServiceRelationState, self).__init__(client)
         self._service_id = service_id
         self._relation_id = relation_id
+        self._relation_scope = relation_scope
         self._role = role
         self._name = name
 
     def __repr__(self):
-        return "<ServiceRelationState name:%s role:%s id:%s>" % (
-            self.relation_name, self.relation_role, self.internal_relation_id)
+        return "<ServiceRelationState name:%s role:%s id:%s scope:%s>" % (
+            self.relation_name,
+            self.relation_role,
+            self.internal_relation_id,
+            self.relation_scope)
 
     @property
     def internal_relation_id(self):
         return self._relation_id
 
+    @staticmethod
+    def get_relation_ident(relation_name, internal_relation_id):
+        """Returns <relation name>:<normalized internal relation id>"""
+
+        # NOTE: ideally this function would not be directly exposed,
+        # however, UnitRelationLifecycle currently uses internal
+        # relation ids, and needs to reconstruct the workflow, which
+        # requires an external relation id, from a file when a
+        # relation has departed.
+
+        # normalize internal ids like 'relation-0000000042' to '42'
+        digits = internal_relation_id.split('-')[1]
+        normalized = digits.lstrip("0")
+        if not normalized:
+            normalized = '0'
+        return "%s:%s" % (relation_name, normalized)
+
+    @property
+    def relation_ident(self):
+        """Returns <relation name>:<normalized internal relation id>"""
+        return self.get_relation_ident(
+            self.relation_name, self.internal_relation_id)
+
     @property
     def internal_service_id(self):
         return self._service_id
@@ -214,38 +271,88 @@ class ServiceRelationState(object):
         """The service's role within the relation."""
         return self._role
 
+    @property
+    def relation_scope(self):
+        """The scope of the relationship."""
+        return self._relation_scope
+
+    def _get_scope_path(self, unit_state, container):
+        relation_scope = self._relation_scope
+        if relation_scope == "container":
+            if container is None:
+                relation_scope_value = unit_state.internal_id
+            else:
+                relation_scope_value = container.internal_id
+        else:
+            relation_scope_value = ""
+
+        scope_path = "/relations/%s/%s" % (
+            self._relation_id, relation_scope_value)
+        return scope_path.rstrip("/")
+
+    @inlineCallbacks
+    def _create_scope_container(self, scope_path, unit_name):
+        # this is a container scoped relation and we must add
+        # node_data as was done in _add_relation_state.
+        # before we can do that however we need to extract the
+        # proper relation information from the topology. This
+        # includes the relation name and the relation_role
+        from .service import parse_service_name
+        topology = yield self._read_topology()
+
+        service_id = topology.find_service_with_name(
+            parse_service_name(unit_name))
+        interface, relation_data = topology.get_relation_service(
+            self._relation_id, service_id)
+
+        try:
+            yield self._client.create(scope_path)
+        except zookeeper.NodeExistsException:
+            pass
+
+        node_data = yaml.safe_dump(relation_data)
+        role_path = "%s/%s" % (scope_path, relation_data["role"])
+
+        yield retry_change(
+            self._client, role_path, lambda c, s: node_data)
+        yield retry_change(
+            self._client, "%s/settings" % scope_path, lambda c, s: "")
+
     @inlineCallbacks
     def add_unit_state(self, unit_state):
         """Add a unit to the service relation.
 
         This api is intended for use by the unit agent, as it
         also creates an ephemeral presence node, denoting the
-        active existance of the unit in the relation.
+        active existence of the unit in the relation.
 
         returns a unit relation state.
         """
-        settings_path = "/relations/%s/settings/%s" % (
-            self._relation_id, unit_state.internal_id)
+        container = yield unit_state.get_container()
+        scope_path = self._get_scope_path(unit_state, container)
+        settings_path = "%s/settings/%s" % (
+            scope_path, unit_state.internal_id)
 
-        # We create settings node first, so that presence node events
-        # have a chance to inspect state.
+        # Pre-populate the relation node with the node's private address.
+        if self._relation_scope == "container":
+            # Create the service relation node data in the proper scope
+            yield self._create_scope_container(
+                scope_path, unit_state.unit_name)
+        if container:
+            private_address = yield container.get_private_address()
+        else:
+            private_address = yield unit_state.get_private_address()
 
-        # Prepopulate the relation node with the node's private address.
-        private_address = yield unit_state.get_private_address()
-        try:
-            yield self._client.create(
-                settings_path,
-                yaml.safe_dump({"private-address": private_address}))
-        except zookeeper.NodeExistsException:
-            # previous persistent settings are not an error, but
-            # update the unit address
-            def update_address(content, stat):
+        def update_address(content, stat):
+            unit_map = None
+            if content:
                 unit_map = yaml.load(content)
-                if not unit_map:
-                    unit_map = {}
-                unit_map["private-address"] = private_address
-                return yaml.safe_dump(unit_map)
-            yield retry_change(self._client, settings_path, update_address)
+            if not unit_map:
+                unit_map = {}
+            unit_map["private-address"] = private_address
+            return yaml.safe_dump(unit_map)
+
+        yield retry_change(self._client, settings_path, update_address)
 
         # Update the unit name -> id mapping on the relation node
         def update_unit_mapping(content, stat):
@@ -253,7 +360,7 @@ class ServiceRelationState(object):
                 unit_map = yaml.load(content)
             else:
                 unit_map = {}
-            # If its already present, we're done, just return the
+            # If it's already present, we're done, just return the
             # existing content, to avoid unstable yaml dict
             # serialization.
             if unit_state.internal_id in unit_map:
@@ -266,8 +373,14 @@ class ServiceRelationState(object):
                            update_unit_mapping)
 
         # Create the presence node.
-        alive_path = "/relations/%s/%s/%s" % (
-            self._relation_id, self._role, unit_state.internal_id)
+        role_path = scope_path + "/" + self._role
+        alive_path = role_path + "/" + unit_state.internal_id
+
+        try:
+            # create the role node
+            yield self._client.create(role_path)
+        except zookeeper.NodeExistsException:
+            pass
 
         try:
             yield self._client.create(alive_path, flags=zookeeper.EPHEMERAL)
@@ -280,13 +393,21 @@ class ServiceRelationState(object):
                 self._client,
                 self._service_id,
                 unit_state.internal_id,
-                self._relation_id))
+                self._relation_id,
+                self._relation_scope))
 
     @inlineCallbacks
     def get_unit_state(self, unit_state):
         """Given a service unit state, return its unit relation state."""
-        alive_path = "/relations/%s/%s/%s" % (
-            self._relation_id, self._role, unit_state.internal_id)
+        if self._relation_scope == "global":
+            alive_path = "/relations/%s/%s/%s" % (
+                self._relation_id, self._role, unit_state.internal_id)
+        else:
+            container = yield unit_state.get_container()
+            scope_path = self._get_scope_path(unit_state, container)
+            alive_path = "%s/%s/%s" % (
+                scope_path, self._role, unit_state.internal_id)
+
         stat = yield self._client.exists(alive_path)
 
         if not stat:
@@ -298,7 +419,8 @@ class ServiceRelationState(object):
                 self._client,
                 self._service_id,
                 unit_state.internal_id,
-                self._relation_id))
+                self._relation_id,
+                self._relation_scope))
 
     @inlineCallbacks
     def get_service_states(self):
@@ -322,11 +444,14 @@ class UnitRelationState(StateBase):
     """A service unit's relation state.
     """
 
-    def __init__(self, client, service_id, unit_id, relation_id):
+    def __init__(
+        self, client, service_id, unit_id, relation_id, relation_scope):
+        #
         super(UnitRelationState, self).__init__(client)
         self._service_id = service_id
         self._unit_id = unit_id
         self._relation_id = relation_id
+        self._relation_scope = relation_scope
 
         # cached value
         self._cached_relation_role = None
@@ -343,6 +468,10 @@ class UnitRelationState(StateBase):
     def internal_relation_id(self):
         return self._relation_id
 
+    @property
+    def relation_scope(self):
+        return self._relation_scope
+
     @inlineCallbacks
     def set_data(self, data):
         """Set the relation local configuration data for a unit.
@@ -350,25 +479,30 @@ class UnitRelationState(StateBase):
         This call overwrites any data currently in the node with the
         dictionary supplied as `data`.
         """
-        unit_settings_path = "/relations/%s/settings/%s" % (
-            self._relation_id, self._unit_id)
-
+        path = yield self.get_settings_path()
         # encode as a YAML string
         data = yaml.safe_dump(data)
 
         yield retry_change(
-            self._client, unit_settings_path,
-            lambda content, stat: data)
+            self._client, path, lambda content, stat: data)
 
     @inlineCallbacks
     def get_data(self):
         """Get the relation local configuration data for a unit.
         """
-        data, stat = yield self._client.get(
-            "/relations/%s/settings/%s" % (
-                self._relation_id, self._unit_id))
+        path = yield self.get_settings_path()
+        data, stat = yield self._client.get(path)
         returnValue(data)
 
+    @inlineCallbacks
+    def get_settings_path(self):
+        parts = ["/relations",
+                 self.internal_relation_id,
+                 (yield self._get_container_id()),
+                 "settings",
+                 self.internal_unit_id]
+        returnValue("/".join(filter(None, parts)))
+
     @inlineCallbacks
     def get_relation_role(self):
         """Return the service's role within the relation.
@@ -379,6 +513,9 @@ class UnitRelationState(StateBase):
         # Perhaps this information could be passed directly into the
         # unit relation state constructor, its already got 4 params though.
         topology = yield self._read_topology()
+        if not (topology.has_relation(self._relation_id) and \
+                topology.has_service(self._service_id)):
+            raise StateChanged()
         relation_type, info = topology.get_relation_service(self._relation_id,
                                                             self._service_id)
         relation_role = info["role"]
@@ -386,26 +523,46 @@ class UnitRelationState(StateBase):
         self._cached_relation_role = relation_role
         returnValue(relation_role)
 
+    @inlineCallbacks
+    def _get_container_id(self):
+        topology = yield self._read_topology()
+        container_id = topology.get_service_unit_container(self._unit_id)
+
+        if container_id is None and self._relation_scope == "global":
+            return
+        elif container_id is None:
+            returnValue(self._unit_id)
+        # Container id comes back as as svc-id, rel-id, unit-id
+        returnValue(container_id[-1])
+
     @inlineCallbacks
     def get_related_unit_container(self):
-        """Return the path to the relation role container of the related units.
+        """Return the path to the relation role container of the related
+        units.
+
+        If container_unit_state is provided it will be used to compute
+        the path into a single container's scope for this
+        relationship.
+
         """
         relation_role = yield self.get_relation_role()
+        endpoint_container = ["/relations", self._relation_id]
+        if self._relation_scope == "container":
+            container_id = yield self._get_container_id()
+            endpoint_container.append(container_id)
         if relation_role == "server":
-            endpoint_container = "/relations/%s/%s" % (self._relation_id,
-                                                       "client")
+            endpoint_container.append("client")
         elif relation_role == "client":
-            endpoint_container = "/relations/%s/%s" % (self._relation_id,
-                                                       "server")
+            endpoint_container.append("server")
         elif relation_role == "peer":
-            endpoint_container = "/relations/%s/peer" % self._relation_id
+            endpoint_container.append("peer")
         else:
             topology = yield self._read_topology()
             service_name = topology.get_service_name(self._service_id)
             raise UnknownRelationRole(
                 self._relation_id, relation_role, service_name)
 
-        returnValue(endpoint_container)
+        returnValue("/".join(endpoint_container))
 
     @inlineCallbacks
     def watch_related_units(self, cb_members, cb_settings):
@@ -433,17 +590,15 @@ class UnitRelationState(StateBase):
         """
         relation_role = yield self.get_relation_role()
         endpoint_container = yield self.get_related_unit_container()
-
         # Determine the watcher implementation.
-        if relation_role == "server":
-            watcher_factory = ClientServerUnitWatcher
-        elif relation_role == "client":
+        if relation_role in ["server", "client"]:
             watcher_factory = ClientServerUnitWatcher
         elif relation_role == "peer":
             watcher_factory = PeerUnitWatcher
 
         watcher = watcher_factory(
-            self._client, self, endpoint_container, cb_members, cb_settings)
+            self._client, self, endpoint_container, cb_members, cb_settings,
+            self._relation_scope)
         returnValue(watcher)
 
 
@@ -475,7 +630,9 @@ class RelationUnitWatcherBase(StateBase):
                  watcher_unit,
                  unit_container_path,
                  cb_members,
-                 cb_settings):
+                 cb_settings,
+                 relation_scope=None):
+
         super(RelationUnitWatcherBase, self).__init__(client)
         self._units = []
         self._watcher_unit = watcher_unit
@@ -484,6 +641,8 @@ class RelationUnitWatcherBase(StateBase):
         self._cb_settings = cb_settings
         self._stopped = None
         self._unit_name_map = None
+        self._relation_scope = relation_scope
+        self._scope_path = dirname(unit_container_path)
         self._log = logging.getLogger("unit.relation.watch")
 
     @property
@@ -561,8 +720,10 @@ class RelationUnitWatcherBase(StateBase):
         corresponding lists of unit names as results.
         """
         if not self._unit_name_map:
-            content, stat = yield self._client.get(
-                dirname(self._container_path))
+            relation_path = dirname(self._container_path)
+            if self._relation_scope == "container":
+                relation_path = dirname(relation_path)
+            content, stat = yield self._client.get(relation_path)
             self._unit_name_map = yaml.load(content)
         results = []
         for unit_id_list in unit_ids:
@@ -643,9 +804,8 @@ class RelationUnitWatcherBase(StateBase):
         settings_watches = []
         # Watch new settings node for changes.
         for unit_id in added:
-            settings_path = "/relations/%s/settings/%s" % (
-                self._watcher_unit.internal_relation_id,
-                unit_id)
+            settings_path = "%s/settings/%s" % (
+                self._scope_path, unit_id)
 
             # Since we have a concurrent execution model, unit tests,
             # will error out since this callback might still be
