@@ -1,21 +1,27 @@
-import yaml
+import logging
+import os
 import time
 
+import yaml
 import zookeeper
-import logging
+
 
 from twisted.internet.defer import (
     inlineCallbacks, returnValue, Deferred, fail, succeed)
-from txzookeeper import ZookeeperClient
 
+from juju.charm.directory import CharmDirectory
 from juju.charm.tests import local_charm_id
-from juju.machine.constraints import Constraints
+from juju.machine.tests.test_constraints import dummy_constraints
+from juju.charm.tests.test_metadata import test_repository_path
+from juju.charm.tests.test_repository import unbundled_repository
+from juju.lib.pick import pick_attr
 from juju.state.charm import CharmStateManager
 from juju.state.endpoint import RelationEndpoint
 from juju.state.errors import (
     DuplicateEndpoints, IncompatibleEndpoints, RelationAlreadyExists,
     RelationStateNotFound, StateChanged, UnitRelationStateNotFound,
-    UnknownRelationRole, ServiceStateNameInUse)
+    UnknownRelationRole, ServiceStateNameInUse, ServiceStateNotFound,
+    CharmStateNotFound)
 
 from juju.state.relation import (
     RelationStateManager, ServiceRelationState, UnitRelationState)
@@ -28,6 +34,7 @@ class RelationTestBase(StateTestBase):
     @inlineCallbacks
     def setUp(self):
         yield super(RelationTestBase, self).setUp()
+        yield self.push_default_config()
         self.relation_manager = RelationStateManager(self.client)
         self.charm_manager = CharmStateManager(self.client)
         self.service_manager = ServiceStateManager(self.client)
@@ -40,11 +47,31 @@ class RelationTestBase(StateTestBase):
                 local_charm_id(self.charm), self.charm, "")
         try:
             service_state = yield self.service_manager.add_service_state(
-                name, self.charm_state, Constraints.from_strs("dummy", []))
+                name, self.charm_state, dummy_constraints)
         except ServiceStateNameInUse:
             service_state = yield self.service_manager.get_service_state(name)
         returnValue(service_state)
 
+    @inlineCallbacks
+    def add_service_from_charm(
+                    self, service_name, charm_id=None, constraints=None,
+                    charm_dir=None, charm_name=None):
+        """Add a service from a charm.
+        """
+        if not charm_id and charm_dir is None:
+            charm_name = charm_name or service_name
+            charm_dir = CharmDirectory(os.path.join(
+                    test_repository_path, "series", charm_name))
+        if charm_id is None:
+            charm_state = yield self.charm_manager.add_charm_state(
+                    local_charm_id(charm_dir), charm_dir, "")
+        else:
+            charm_state = yield self.charm_manager.get_charm_state(
+                    charm_id)
+        service_state = yield self.service_manager.add_service_state(
+                service_name, charm_state, constraints or dummy_constraints)
+        returnValue(service_state)
+
     @inlineCallbacks
     def add_relation(self, relation_type, *services):
         """Support older tests that don't use `RelationEndpoint`s"""
@@ -86,8 +113,7 @@ class RelationTestBase(StateTestBase):
         # 2. And join together in a relation
         relation_state, service_relation_states = \
                         yield self.relation_manager.add_relation_state(
-            *endpoints)
-
+                            *endpoints)
         # 3. Add a service unit to only the first endpoint - we need
         #    to test what happens when service units are added to the
         #    other service state (if any), so do so separately
@@ -238,6 +264,88 @@ class RelationTestBase(StateTestBase):
             state_dict["unit"].internal_id)
         return unit_relation_path
 
+    @inlineCallbacks
+    def get_local_charm(self, charm_name):
+        charm_dir = CharmDirectory(
+            os.path.join(unbundled_repository, "series", charm_name))
+        try:
+            charm_state = yield self.charm_manager.get_charm_state(
+                local_charm_id(charm_dir))
+        except CharmStateNotFound:
+            charm_state = yield self.charm_manager.add_charm_state(
+                local_charm_id(charm_dir), charm_dir, "")
+        returnValue(charm_state)
+
+    @inlineCallbacks
+    def get_subordinate_charm(self):
+        """Return charm state for a subordinate charm.
+
+        Many tests rely on adding relationships to a proper subordinate.
+        This return the charm state of a testing subordinate charm.
+        """
+        sub_charm = yield self.get_local_charm("logging")
+        returnValue(sub_charm)
+
+    @inlineCallbacks
+    def get_service_and_units_by_charm(self, charm_state,
+                                       units=None,
+                                       containers=None,
+                                       service_name=None):
+        """Return [service_state, [o..n units]]
+
+        `units (int)` is provided that many units will be added
+
+        `containers` if provided it should be a list of unit states for
+        containers that should be associated with each new unit (in order).
+        This option implies units == len(containers) (though this is checked).
+
+        `service_name` optional name to use for service, defaults to charm
+        name.
+
+        """
+        if not service_name:
+            service_name = charm_state.name
+
+        try:
+            service_state = yield self.service_manager.get_service_state(
+                service_name)
+
+        except ServiceStateNotFound:
+            service_state = yield self.service_manager.add_service_state(
+                service_name, charm_state, dummy_constraints)
+
+        if containers:
+            if units is None:
+                units = len(containers)
+            elif len(containers) != units:
+                raise ValueError(
+                    "Containers and number of expected units mismatch")
+
+        unit_states = []
+        if units:
+            for i in range(units):
+                container = None
+                if containers:
+                    container = containers[i]
+                unit = yield service_state.add_unit_state(
+                    container=container)
+                unit_states.append(unit)
+
+        returnValue([service_state, unit_states])
+
+    @inlineCallbacks
+    def get_service_and_units_by_charm_name(self, charm_name,
+                                            units=None,
+                                            containers=None,
+                                            service_name=None):
+        charm_state = yield self.get_local_charm(charm_name)
+        returnValue((
+            yield self.get_service_and_units_by_charm(
+                charm_state,
+                units=units,
+                containers=containers,
+                service_name=service_name)))
+
 
 class RelationStateManagerTest(RelationTestBase):
 
@@ -262,7 +370,7 @@ class RelationStateManagerTest(RelationTestBase):
         """Test adding a relation to a nonexistent service"""
         mysql_ep = RelationEndpoint("mysql", "mysql", "db", "server")
         blog_ep = RelationEndpoint("wordpress", "mysql", "mysql", "client")
-        yield self.add_service("mysql")
+        yield self.add_service("wordpress")
         # but didn't create the service for wordpress
         yield self.assertFailure(
             self.relation_manager.add_relation_state(
@@ -372,6 +480,58 @@ class RelationStateManagerTest(RelationTestBase):
             self.relation_manager.add_relation_state(mysql_ep, drizzle_ep),
             IncompatibleEndpoints)
 
+    @inlineCallbacks
+    def test_add_relation_state_scope_container_relation(self):
+        """Verify that container scope is applied to relation.
+
+        Even in the case where only one endpoint is marked as scope:container.
+        """
+        mysql_ep = RelationEndpoint(
+            "mysql", "juju-info", "juju-info", "server", "global")
+        logging_ep = RelationEndpoint(
+            "logging", "juju-info", "juju-info", "client", "container")
+        yield self.add_service("mysql")
+        yield self.add_service("logging")
+        relation_state, service_states = \
+            yield self.relation_manager.add_relation_state(mysql_ep,
+                                                           logging_ep)
+
+        # verify that the relation state instances are both marked
+        # with the live scope (container). This happens even though
+        # the provides side of the relation is global
+        for service_state in service_states:
+            self.assertEqual(service_state.relation_scope, "container")
+
+    @inlineCallbacks
+    def test_add_relation_state_scope_container_relation_2_containers(self):
+        mysql_ep = RelationEndpoint(
+            "mysql", "juju-info", "juju-info", "server", "container")
+        logging_ep = RelationEndpoint(
+            "logging", "juju-info", "juju-info", "client", "container")
+        yield self.add_service("mysql")
+        yield self.add_service("logging")
+        relation_state, service_states = \
+            yield self.relation_manager.add_relation_state(mysql_ep,
+                                                           logging_ep)
+
+        for service_state in service_states:
+            self.assertEqual(service_state.relation_scope, "container")
+
+    @inlineCallbacks
+    def test_add_relation_state_scope_container_scoped_principal(self):
+        mysql_ep = RelationEndpoint(
+            "mysql", "juju-info", "juju-info", "server", "container")
+        logging_ep = RelationEndpoint(
+            "logging", "juju-info", "juju-info", "client", "global")
+        yield self.add_service("mysql")
+        yield self.add_service("logging")
+        relation_state, service_states = \
+            yield self.relation_manager.add_relation_state(mysql_ep,
+                                                           logging_ep)
+
+        for service_state in service_states:
+            self.assertEqual(service_state.relation_scope, "container")
+
     @inlineCallbacks
     def test_remove_relation_state(self):
         """Removing a relation will remove it from the topology."""
@@ -529,11 +689,17 @@ class ServiceRelationStateTest(RelationTestBase):
             self.service_state1)
         self.service1_relation = relations.pop()
 
-    def get_presence_path(self, relation_state, relation_role, unit_state):
-        presence_path = "/relations/%s/%s/%s" % (
+    def get_presence_path(
+        self, relation_state, relation_role, unit_state, container=None):
+        #
+        container = container.internal_id if container else None
+
+        presence_path = "/".join(filter(None, [
+            "/relations",
             relation_state.internal_id,
+            container,
             relation_role,
-            unit_state.internal_id)
+            unit_state.internal_id]))
         return presence_path
 
     def test_property_internal_service_id(self):
@@ -551,7 +717,8 @@ class ServiceRelationStateTest(RelationTestBase):
         id = "relation-0000000000"
         self.assertEqual(
             repr(self.service1_relation),
-            "<ServiceRelationState name:dev-connect role:prod id:%s>" % id)
+            "<ServiceRelationState name:dev-connect "
+            "role:prod id:%s scope:global>" % id)
 
     def test_property_relation_name(self):
         """
@@ -560,6 +727,40 @@ class ServiceRelationStateTest(RelationTestBase):
         """
         self.assertEqual(self.service1_relation.relation_name, "dev-connect")
 
+    @inlineCallbacks
+    def assert_relation_idents(self, service, expected):
+        relations = yield self.relation_manager.get_relations_for_service(
+            service)
+        self.assertEqual(set((r.relation_ident for r in relations)),
+                         set(expected))
+
+    @inlineCallbacks
+    def test_property_relation_id(self):
+        """Verify normalization of relation id and
+           correct selection of relation name.
+        """
+        yield self.assert_relation_idents(
+            self.service_state1, ["dev-connect:0"])
+        yield self.assert_relation_idents(
+            self.service_state2, ["prod-connect:0"])
+
+        # Setup another group of services and establish relations to
+        # verify working with non-zero id and with multiple consumers
+        mysql = yield self.add_service("mysql")
+        blog1 = yield self.add_service("blog1")
+        blog2 = yield self.add_service("blog2")
+        yield self.add_relation(
+            "mysql",
+            (mysql, "database", "server"),
+            (blog1, "db", "client"))
+        yield self.add_relation(
+            "mysql",
+            (mysql, "database", "server"),
+            (blog2, "db", "client"))
+        yield self.assert_relation_idents(blog1, ["db:1"])
+        yield self.assert_relation_idents(blog2, ["db:2"])
+        yield self.assert_relation_idents(mysql, ["database:1", "database:2"])
+
     @inlineCallbacks
     def test_add_unit_state(self):
         """The service state is used to create units in the relation."""
@@ -590,7 +791,6 @@ class ServiceRelationStateTest(RelationTestBase):
 
         # add the unit agent
         yield self.service1_relation.add_unit_state(unit_state)
-
         # wait for the watches
         yield state_created
 
@@ -612,6 +812,61 @@ class ServiceRelationStateTest(RelationTestBase):
         self.assertEqual(
             yaml.load(content), {"private-address": "foobar.local"})
 
+    @inlineCallbacks
+    def test_add_unit_state_scope_container_relation(self):
+        """Verify that container scope is applied to relation.
+
+        Even in the case where only one endpoint is marked as scope:container.
+        """
+        mysql_ep = RelationEndpoint("mysql", "juju-info", "juju-info",
+                                    "server", "global")
+        logging_ep = RelationEndpoint("logging", "juju-info", "juju-info",
+                                      "client", "container")
+
+        mysql = yield self.add_service("mysql")
+        log_charm = yield self.get_subordinate_charm()
+        log = yield self.service_manager.add_service_state("logging",
+                                                           log_charm,
+                                                           dummy_constraints)
+        self.assertTrue(log.is_subordinate())
+
+        relation_state, service_states = (yield
+            self.relation_manager.add_relation_state(
+                mysql_ep, logging_ep))
+
+        mu1 = yield mysql.add_unit_state()
+        lu1 = yield log.add_unit_state(mu1)
+
+        mystate = pick_attr(service_states, relation_role="server")
+        logstate = pick_attr(service_states, relation_role="client")
+
+        lurs1 = yield logstate.add_unit_state(lu1)
+        yield lurs1.set_data(dict(magic="horse"))
+        murs1 = yield mystate.add_unit_state(mu1)
+        yield murs1.set_data(dict(magic="unicorn"))
+
+        # verify that the relation state instances are both marked
+        # with the live scope (container). This happens even though
+        # the provides side of the relation is global
+        for service_state in service_states:
+            self.assertEqual(service_state.relation_scope, "container")
+
+        yield self.assertTree("/relations/%s/%s" % (
+            murs1.internal_relation_id, murs1.internal_unit_id),
+            {"unit-0000000000": {
+                "client": {"contents": {"name": "juju-info",
+                                        "role": "client"},
+                           "unit-0000000002": {"contents": {}}},
+                "contents": {},
+                "server": {"contents": {"name": "juju-info",
+                                        "role": "server"},
+                           "unit-0000000000": {"contents": {}}},
+                "settings": {"contents": {},
+                             "unit-0000000000": {
+                                 "contents": {"private-address": None}},
+                             "unit-0000000002": {
+                                 "contents": {"private-address": None}}}}})
+
     @inlineCallbacks
     def test_presence_node_is_ephemeral(self):
         """
@@ -623,12 +878,13 @@ class ServiceRelationStateTest(RelationTestBase):
 
         # manually construct a unit relation state using a separate
         # connection.
-        client2 = ZookeeperClient(self.client.servers)
+        client2 = self.get_zookeeper_client()
         yield client2.connect()
         service_relation = ServiceRelationState(
             client2,
             self.service_state1.internal_id,
             self.relation_state.internal_id,
+            "global",
             "prod",
             "name")
         yield service_relation.add_unit_state(unit_state)
@@ -721,6 +977,16 @@ class ServiceRelationStateTest(RelationTestBase):
                          set((self.service_state1,
                               self.service_state2)))
 
+    def test_get_relation_ident(self):
+        self.assertEqual(
+            ServiceRelationState.get_relation_ident(
+                "db", "relation-0000000042"),
+            "db:42")
+        self.assertEqual(
+            ServiceRelationState.get_relation_ident(
+                "db", "relation-0000000000"),
+            "db:0")
+
 
 class WatchChecker(object):
     """Helper class to simplify UnitRelationState.watch_related_units tests.
@@ -854,6 +1120,24 @@ class UnitRelationStateTest(RelationTestBase):
         role = yield states["unit_relation"].get_relation_role()
         self.assertEqual("server", role)
 
+    @inlineCallbacks
+    def test_get_relation_role_on_removed_relation(self):
+        """Verify `StateChanged` raised if relation is removed."""
+        states = yield self.add_relation_service_unit(
+            "webcache", "varnish", "name", "server")
+        yield self.relation_manager.remove_relation_state(states["relation"])
+        yield self.assertFailure(
+            states["unit_relation"].get_relation_role(), StateChanged)
+
+    @inlineCallbacks
+    def test_get_relation_role_on_removed_service(self):
+        """Verify `StateChanged` raised if service is removed."""
+        states = yield self.add_relation_service_unit(
+            "webcache", "varnish", "name", "server")
+        yield self.service_manager.remove_service_state(states["service"])
+        yield self.assertFailure(
+            states["unit_relation"].get_relation_role(), StateChanged)
+        
     @inlineCallbacks
     def test_get_related_unit_container(self):
         """Retrieve the container path of the related units."""
@@ -1012,10 +1296,12 @@ class UnitRelationStateTest(RelationTestBase):
             wordpress_ep, mysql_ep)
         mysql_states = yield self.add_opposite_service_unit(
             wordpress_states)
+
         mysql_unit = mysql_states["unit"]
 
         checker = WatchChecker(self)
         watcher = yield checker.watch(wordpress_states["unit_relation"])
+
         yield watcher.start()
         yield checker.assert_members_cb(0, [], [mysql_unit])
         yield checker.assert_settings_cb(1, mysql_unit, 0)
@@ -1044,8 +1330,10 @@ class UnitRelationStateTest(RelationTestBase):
 
         checker = WatchChecker(self)
         watcher = yield checker.watch(mysql_states["unit_relation"])
+
         yield watcher.start()
         yield self.poke_zk()
+
         checker.assert_cb_count(0)
 
         # add the client service and a unit of that
@@ -1055,6 +1343,66 @@ class UnitRelationStateTest(RelationTestBase):
         yield checker.assert_members_cb(0, [], [wordpress_unit])
         yield checker.assert_settings_cb(1, wordpress_unit, 0)
 
+    @inlineCallbacks
+    def test_watch_server_client_with_new_subordinate_service(self):
+        """We simulate a server watching a client.
+        """
+        mysql_ep = RelationEndpoint("mysql", "juju-info", "juju-info",
+                                    "server", "global")
+        logging_ep = RelationEndpoint("logging", "juju-info", "juju-info",
+                                      "client", "container")
+
+        mysql, my_units = yield self.get_service_and_units_by_charm_name(
+            "mysql", 2)
+        self.assertFalse((yield mysql.is_subordinate()))
+
+        log, log_units = yield self.get_service_and_units_by_charm_name(
+            "logging")
+        self.assertTrue((yield log.is_subordinate()))
+
+        # add the relationship so we can create units with  containers
+        relation_state, service_states = (yield
+            self.relation_manager.add_relation_state(
+            mysql_ep, logging_ep))
+
+        log, log_units = yield self.get_service_and_units_by_charm_name(
+            "logging",
+            containers=my_units)
+        self.assertTrue((yield log.is_subordinate()))
+        for lu in log_units:
+            self.assertTrue((yield lu.is_subordinate()))
+
+        mu1, mu2 = my_units
+        lu1, lu2 = log_units
+
+        mystate = pick_attr(service_states, relation_role="server")
+        logstate = pick_attr(service_states, relation_role="client")
+
+        murs1 = yield mystate.add_unit_state(mu1)
+        lurs1 = yield logstate.add_unit_state(lu1)
+        # add the second container
+        murs2 = yield mystate.add_unit_state(mu2)
+        lurs2 = yield logstate.add_unit_state(lu2)
+
+        @inlineCallbacks
+        def verify_watch(urs, expected):
+            checker = WatchChecker(self)
+            watcher = yield checker.watch(urs)
+
+            self.assertFalse(watcher.running)
+            yield watcher.start()
+            self.assertTrue(watcher.running)
+            # Here we show the watchers bound to a given unit
+            # only see the contained unit
+            yield checker.assert_members_cb(0, [], expected)
+            yield checker.assert_settings_cb(1, expected[0], 0)
+
+        yield verify_watch(murs1, [lu1])
+        yield verify_watch(murs2, [lu2])
+
+        yield verify_watch(lurs1, [mu1])
+        yield verify_watch(lurs2, [mu2])
+
     @inlineCallbacks
     def test_watch_peer(self):
         """Peer relations always watch the peer container.
@@ -1069,6 +1417,7 @@ class UnitRelationStateTest(RelationTestBase):
 
         checker = WatchChecker(self)
         watcher = yield checker.watch(riak_states["unit_relation"])
+
         yield watcher.start()
 
         # wait for initial callbacks
@@ -1137,6 +1486,7 @@ class UnitRelationStateTest(RelationTestBase):
         checker = WatchChecker(self)
         watcher = yield checker.watch(mysql_states["unit_relation"])
         yield watcher.start()
+
         yield checker.assert_members_cb(0, [], [wordpress_unit])
 
     @inlineCallbacks
@@ -1200,6 +1550,7 @@ class UnitRelationStateTest(RelationTestBase):
 
         checker = WatchChecker(self)
         watcher = yield checker.watch(mysql_states["unit_relation"])
+
         yield watcher.start()
         yield checker.assert_members_cb(0, [], [wordpress_unit])
         yield checker.assert_settings_cb(1, wordpress_unit, 0)
@@ -1250,6 +1601,7 @@ class UnitRelationStateTest(RelationTestBase):
 
         checker = WatchChecker(self)
         watcher = yield checker.watch(mysql_states["unit_relation"])
+
         yield watcher.start()
         self.assertTrue(watcher.running)
         yield checker.assert_members_cb(0, [], [wordpress_unit])
@@ -1294,11 +1646,13 @@ class UnitRelationStateTest(RelationTestBase):
             "wordpress", "client-server", "", "client")
         mysql_ep = RelationEndpoint(
             "mysql", "client-server", "", "server")
+
         mysql_states = yield self.add_relation_service_unit_from_endpoints(
             mysql_ep, wordpress_ep)
 
         checker = WatchChecker(self)
         watcher = yield checker.watch(mysql_states["unit_relation"])
+
         yield watcher.start()
         self.assertTrue(watcher.running)
         watcher.stop()
@@ -1428,6 +1782,7 @@ class UnitRelationStateTest(RelationTestBase):
 
         checker = WatchChecker(self, block_cbs=True)
         watcher = yield checker.watch(riak_states["unit_relation"])
+
         yield watcher.start()
 
         # Create a new unit and add it to the relation.
