@@ -8,7 +8,6 @@ from Queue import Queue
 from twisted.internet import reactor
 from twisted.internet.defer import inlineCallbacks
 from twisted.internet.threads import deferToThread
-from txzookeeper import ZookeeperClient
 import yaml
 
 from juju.lib.testing import TestCase
@@ -18,8 +17,6 @@ from juju.state.utils import (
     PortWatcher, remove_tree, dict_merge, get_open_port,
     YAMLState, YAMLStateNodeMixin, AddedItem, ModifiedItem, DeletedItem)
 
-from juju.tests.common import get_test_zookeeper_address
-
 
 class PortWait(TestCase):
 
@@ -193,7 +190,7 @@ class RemoveTreeTest(TestCase):
     def setUp(self):
         yield super(RemoveTreeTest, self).setUp()
         zookeeper.set_debug_level(0)
-        self.client = ZookeeperClient(get_test_zookeeper_address())
+        self.client = self.get_zookeeper_client()
         yield self.client.connect()
 
     @inlineCallbacks
@@ -309,7 +306,7 @@ class YAMLStateTest(TestCase):
     @inlineCallbacks
     def setUp(self):
         zookeeper.set_debug_level(0)
-        self.client = ZookeeperClient(get_test_zookeeper_address())
+        self.client = self.get_zookeeper_client()
         yield self.client.connect()
         self.path = "/zoo"
 
@@ -653,7 +650,7 @@ class YAMLStateNodeMixinTest(TestCase):
     @inlineCallbacks
     def setUp(self):
         zookeeper.set_debug_level(0)
-        self.client = ZookeeperClient(get_test_zookeeper_address())
+        self.client = self.get_zookeeper_client()
         yield self.client.connect()
         self.path = "/path"
 
