@@ -13,9 +13,11 @@ from juju.hooks.scheduler import HookScheduler
 from juju.state.hook import (
     DepartedRelationHookContext, HookContext, RelationChange)
 from juju.state.errors import StopWatcher, UnitRelationStateNotFound
-from juju.state.relation import RelationStateManager, UnitRelationState
+from juju.state.relation import (
+    RelationStateManager, UnitRelationState)
 
 from juju.unit.charm import download_charm
+from juju.unit.deploy import UnitDeployer
 from juju.unit.workflow import RelationWorkflowState
 
 
@@ -246,15 +248,21 @@ class UnitLifecycle(object):
         self._log.debug("configured unit")
 
     @inlineCallbacks
-    def upgrade_charm(self, fire_hooks=True):
+    def upgrade_charm(self, fire_hooks=True, force=False):
         """Upgrade the charm and invoke the upgrade-charm hook if requested.
 
         :param fire_hooks: if False, *and* the actual upgrade operation is not
             necessary, skip the upgrade-charm hook. When the actual charm has
             changed during this invocation, this flag is ignored: hooks will
             always be fired.
+
+        :param force: Boolean, if true then we're merely putting the charm into
+            place on disk, not executing charm hooks.
         """
-        self._log.debug("Upgrading charm")
+        msg = "Upgrading charm"
+        if force:
+            msg += " - forced"
+        self._log.debug(msg)
         upgrade = _CharmUpgradeOperation(
             self._client, self._service, self._unit, self._unit_dir)
         yield self._run_lock.acquire()
@@ -268,9 +276,9 @@ class UnitLifecycle(object):
 
             if upgrade.ready:
                 yield upgrade.run()
-                # We changed the charm just now: we *must* fire hooks.
                 fire_hooks = True
-            if fire_hooks:
+
+            if fire_hooks and not force:
                 yield self._execute_hook("upgrade-charm", now=True)
 
             # Always restart executor on success; charm upgrade operations and
@@ -317,8 +325,8 @@ class UnitLifecycle(object):
             yield self._unit.clear_relation_resolved()
 
         keys = set(relation_resolved).intersection(self._relations)
-        for rel_id in keys:
-            workflow = self._relations[rel_id]
+        for internal_rel_id in keys:
+            workflow = self._relations[internal_rel_id]
             with (yield workflow.lock()):
                 state = yield workflow.get_state()
                 if state != "up":
@@ -360,8 +368,16 @@ class UnitLifecycle(object):
         new_relations = dict(
             (service_relation.internal_relation_id, service_relation)
             for service_relation in new_relations)
+
+        if old_relations:
+            old_relations = dict(
+                (service_relation.internal_relation_id, service_relation)
+                for service_relation in old_relations)
+
         added = set(new_relations.keys()) - set(self._relations.keys())
         removed = set(self._relations.keys()) - set(new_relations.keys())
+        # Could this service be a principal container?
+        is_principal = not (yield self._service.is_subordinate())
 
         # Once we know a relation is departed, *immediately* stop running
         # its hooks. We can't really handle the case in which a hook is
@@ -392,7 +408,9 @@ class UnitLifecycle(object):
         for relation_id in added:
             service_relation = new_relations[relation_id]
             yield self._add_relation(service_relation)
-            self._store_relations()
+            if (is_principal and service_relation.relation_scope == "container"):
+                self._add_subordinate_unit(service_relation)
+            yield self._store_relations()
 
     @inlineCallbacks
     def _add_relation(self, service_relation):
@@ -407,17 +425,57 @@ class UnitLifecycle(object):
 
         lifecycle = UnitRelationLifecycle(
             self._client, self._unit.unit_name, unit_relation,
-            service_relation.relation_name, self._unit_dir, self._state_dir,
-            self._executor)
+            service_relation.relation_ident,
+            self._unit_dir, self._state_dir, self._executor)
 
         workflow = RelationWorkflowState(
             self._client, unit_relation, service_relation.relation_name,
             lifecycle, self._state_dir)
 
         self._relations[service_relation.internal_relation_id] = workflow
+
         with (yield workflow.lock()):
             yield workflow.synchronize()
 
+    @inlineCallbacks
+    def _do_unit_deploy(self, unit_name, machine_id, charm_dir):
+        # this method exists to aid testing rather than being an
+        # inline
+        unit_deployer = UnitDeployer(self._client, machine_id, charm_dir)
+        yield unit_deployer.start("subordinate")
+        yield unit_deployer.start_service_unit(unit_name)
+
+    @inlineCallbacks
+    def _add_subordinate_unit(self, service_relation):
+        """Deploy a subordinate unit for service_relation remote endpoint."""
+        # Figure out the remote service state
+        service_states = yield service_relation.get_service_states()
+        subordinate_service = [s for s in service_states if
+                               s.service_name != self._unit.service_name][0]
+
+        # add a unit state to service (using self._unit as the
+        # principal container)
+        subordinate_unit = yield subordinate_service.add_unit_state(
+            container=self._unit)
+        machine_id = yield self._unit.get_assigned_machine_id()
+
+        subordinate_unit_dir = os.path.dirname(self._unit_dir)
+        charm_dir = os.path.join(subordinate_unit_dir,
+                                 subordinate_unit.unit_name.replace(
+                                     "/", "-"))
+        state_dir = os.path.join(charm_dir, "state")
+        if not os.path.exists(state_dir):
+                os.makedirs(state_dir)
+
+        self._log.debug("deploying %s as subordinate of %s",
+                        subordinate_unit.unit_name,
+                        self._unit.unit_name)
+        # with the relation in place and the units added to the
+        # container we can start the unit agent
+        yield self._do_unit_deploy(subordinate_unit.unit_name,
+                                   machine_id,
+                                   charm_dir)
+
     @property
     def _known_relations_path(self):
         return os.path.join(
@@ -432,10 +490,12 @@ class UnitLifecycle(object):
         can still depart broken relations even if they break while we're not
         running.
         """
-        state = yaml.dump(dict(
-            (relation_id, workflow.relation_name)
-            for (relation_id, workflow) in self._relations.items()))
+        state_dict = {}
+        for relation_wf in self._relations.itervalues():
+            state_dict.update(relation_wf.get_relation_info())
+        state = yaml.dump(state_dict)
         temp_path = self._known_relations_path + "~"
+
         with open(temp_path, "w") as f:
             f.write(state)
         os.rename(temp_path, self._known_relations_path)
@@ -456,13 +516,12 @@ class UnitLifecycle(object):
 
         rsm = RelationStateManager(self._client)
         relations = yield rsm.get_relations_for_service(self._service)
-        relations_by_id = dict(
-            (r.internal_relation_id, r) for r in relations)
+        relations_by_id = dict((r.internal_relation_id, r) for r in relations)
 
         with open(self._known_relations_path) as f:
             known_relations = yaml.load(f.read())
 
-        for relation_id, relation_name in known_relations.items():
+        for relation_id, relation_info in known_relations.items():
             if relation_id in relations_by_id:
                 # The service relation's still around: set up workflow as usual
                 yield self._add_relation(relations_by_id[relation_id])
@@ -471,10 +530,12 @@ class UnitLifecycle(object):
                 # workflow and place it in relations for detection and
                 # removal (with hook-firing) in _process_service_changes.
                 workflow = self._reconstruct_workflow(
-                    relation_id, relation_name)
+                    relation_id,
+                    relation_info["relation_name"],
+                    relation_info["relation_scope"])
                 self._relations[relation_id] = workflow
 
-    def _reconstruct_workflow(self, relation_id, relation_name):
+    def _reconstruct_workflow(self, relation_id, relation_ident, relation_scope):
         """Create a RelationWorkflowState which may refer to outdated state.
 
         This means that *if* this service has already departed the relevant
@@ -486,10 +547,11 @@ class UnitLifecycle(object):
         """
         unit_relation = UnitRelationState(
             self._client, self._service.internal_id, self._unit.internal_id,
-            relation_id)
+            relation_id, relation_scope)
         lifecycle = UnitRelationLifecycle(
-            self._client, self._unit.unit_name, unit_relation, relation_name,
+            self._client, self._unit.unit_name, unit_relation, relation_ident,
             self._unit_dir, self._state_dir, self._executor)
+        relation_name = relation_ident.split(":")[0]
         return RelationWorkflowState(
             self._client, unit_relation, relation_name, lifecycle,
             self._state_dir)
@@ -506,6 +568,7 @@ class UnitLifecycle(object):
         invoker = Invoker(
             HookContext(self._client, self._unit.unit_name), None,
             _EVIL_CONSTANT, socket_path, self._unit_dir, hook_log)
+        yield invoker.start()
 
         if now:
             yield self._executor.run_priority_hook(invoker, hook_path)
@@ -520,6 +583,7 @@ class RelationInvoker(Invoker):
     def get_environment_from_change(self, env, change):
         """Populate environment with relation change information."""
         env["JUJU_RELATION"] = change.relation_name
+        env["JUJU_RELATION_ID"] = change.relation_ident
         env["JUJU_REMOTE_UNIT"] = change.unit_name
         return env
 
@@ -545,11 +609,12 @@ class UnitRelationLifecycle(object):
     discussion of some of the more interesting implementation decisions.
     """
 
-    def __init__(self, client, unit_name, unit_relation, relation_name,
+    def __init__(self, client, unit_name, unit_relation, relation_ident,
                  unit_dir, state_dir, executor):
         self._client = client
         self._unit_dir = unit_dir
-        self._relation_name = relation_name
+        self._relation_ident = relation_ident
+        self._relation_name = relation_ident.split(":")[0]
         self._unit_relation = unit_relation
         self._unit_name = unit_name
         self._executor = executor
@@ -561,7 +626,7 @@ class UnitRelationLifecycle(object):
             state_dir, "%s.schedule" % unit_relation.internal_relation_id)
         self._scheduler = HookScheduler(
             client, self._execute_change_hook, self._unit_relation,
-            self._relation_name, unit_name, schedule_path)
+            self._relation_ident, unit_name, schedule_path)
         self._watcher = None
 
     @property
@@ -633,11 +698,10 @@ class UnitRelationLifecycle(object):
         """
         self._log.debug("depart relation lifecycle")
         unit_id = self._unit_relation.internal_unit_id
-        relation_id = self._unit_relation.internal_relation_id
         context = DepartedRelationHookContext(
             self._client, self._unit_name, unit_id, self._relation_name,
-            relation_id)
-        change = RelationChange(self._relation_name, "departed", "")
+            self._unit_relation.internal_relation_id)
+        change = RelationChange(self._relation_ident, "departed", "")
         invoker = self._get_invoker(context, change)
         hook_name = "%s-relation-broken" % self._relation_name
         yield self._execute_hook(invoker, hook_name, change)
@@ -648,7 +712,6 @@ class UnitRelationLifecycle(object):
             context, change, "constant", socket_path, self._unit_dir,
             hook_log)
 
-    @inlineCallbacks
     def _execute_change_hook(self, context, change):
         """Invoked by the contained HookScheduler, to execute a hook.
 
@@ -657,22 +720,18 @@ class UnitRelationLifecycle(object):
         is specified see ``set_hook_error_handler``.
         """
         if change.change_type == "departed":
-            hook_names = [
-                "%s-relation-departed" % self._relation_name]
+            hook_name = "%s-relation-departed" % self._relation_name
         elif change.change_type == "joined":
-            hook_names = [
-                "%s-relation-joined" % self._relation_name,
-                "%s-relation-changed" % self._relation_name]
+            hook_name = "%s-relation-joined" % self._relation_name
         else:
-            hook_names = ["%s-relation-changed" % self._relation_name]
+            hook_name = "%s-relation-changed" % self._relation_name
 
         invoker = self._get_invoker(context, change)
-        for hook_name in hook_names:
-            if not (yield self._execute_hook(invoker, hook_name, change)):
-                break
+        return self._execute_hook(invoker, hook_name, change)
 
     @inlineCallbacks
     def _execute_hook(self, invoker, hook_name, change):
+        yield invoker.start()
         hook_path = os.path.join(
             self._unit_dir, "charm", "hooks", hook_name)
         yield self._run_lock.acquire()
