@@ -1,21 +1,30 @@
+import StringIO
+import itertools
 import logging
 import os
-import sys
-import StringIO
+import shutil
 import stat
+import sys
 
 import yaml
 import zookeeper
 
-from twisted.internet.defer import inlineCallbacks, Deferred, fail, returnValue
+from twisted.internet.defer import (inlineCallbacks, Deferred,
+                                    fail, returnValue)
 
+from juju.charm.directory import CharmDirectory
+from juju.charm.tests.test_repository import unbundled_repository
 from juju.charm.url import CharmURL
+
 from juju.control.tests.test_upgrade_charm import CharmUpgradeTestBase
 from juju.errors import CharmInvocationError, CharmError, CharmUpgradeError
 from juju.hooks.invoker import Invoker
 from juju.hooks.executor import HookExecutor
+from juju.machine.tests.test_constraints import series_constraints
+
 from juju.state.endpoint import RelationEndpoint
 from juju.state.errors import UnitRelationStateNotFound
+from juju.state.machine import MachineStateManager
 from juju.state.relation import ClientServerUnitWatcher
 from juju.state.service import NO_HOOKS
 from juju.state.tests.test_relation import RelationTestBase
@@ -48,7 +57,6 @@ class LifecycleTestBase(RelationTestBase):
     @inlineCallbacks
     def setUp(self):
         yield super(LifecycleTestBase, self).setUp()
-
         if self.juju_directory is None:
             self.juju_directory = self.makeDir()
 
@@ -145,7 +153,7 @@ class LifecycleTestBase(RelationTestBase):
         workflow.set_observer(observer)
         return state_changed
 
-    def capture_output(self, stdout=True):
+    def capture_output(self, stdout=True, channels=()):
         """Convience method to capture log output.
 
         Useful tool for observing interaction between components.
@@ -154,13 +162,17 @@ class LifecycleTestBase(RelationTestBase):
             output = sys.stdout
         else:
             output = StringIO.StringIO()
-        for log_name, indent in (
+        channels = zip(channels, itertools.repeat(0))
+
+        for log_name, indent in itertools.chain((
             ("statemachine", 0),
             ("hook.executor", 2),
             ("hook.scheduler", 1),
+            ("unit.deploy", 2),
             ("unit.lifecycle", 1),
             ("unit.relation.watch", 1),
-            ("unit.relation.lifecycle", 1)):
+            ("unit.relation.lifecycle", 1)),
+            channels):
             formatter = logging.Formatter(
                 (" " * indent) + "%(name)s: %(message)s")
             self.capture_logging(
@@ -346,7 +358,6 @@ class LifecycleResolvedTest(LifecycleTestBase):
         yield self.states["unit"].set_relation_resolved(
             {self.states["unit_relation"].internal_relation_id: NO_HOOKS})
 
-
         # Wait for the relation to come back up
         value = yield self.states["unit"].get_relation_resolved()
 
@@ -370,6 +381,12 @@ class UnitLifecycleTest(LifecycleTestBase):
             self.client, self.states["unit"], self.states["service"],
             self.unit_directory, self.state_directory, self.executor)
 
+    @inlineCallbacks
+    def tearDown(self):
+        if self.lifecycle.running:
+            yield self.lifecycle.stop(fire_hooks=False)
+        yield super(UnitLifecycleTest, self).tearDown()
+
     @inlineCallbacks
     def test_hook_invocation(self):
         """Verify lifecycle methods invoke corresponding charm hooks.
@@ -525,7 +542,6 @@ class UnitLifecycleTest(LifecycleTestBase):
             sequence=[
                 "app-relation-joined",   # joined event fires join hook,
                 "app-relation-changed",  # followed by changed hook
-                "app-relation-changed",
                 "app-relation-departed"])
 
         # add another.
@@ -598,7 +614,7 @@ class UnitLifecycleTest(LifecycleTestBase):
 
     @inlineCallbacks
     def test_lifecycle_start_stop_starts_relations(self):
-        """Starting a stopped lifecycle, restarts relation events.
+        """Starting a stopped lifecycle will restart relation events.
         """
         wordpress1_states = yield self.add_opposite_service_unit(self.states)
         wordpress2_states = yield self.add_opposite_service_unit(
@@ -606,11 +622,15 @@ class UnitLifecycleTest(LifecycleTestBase):
                 self.states,
                 RelationEndpoint(
                     "wordpress-2", "client-server", "db", "client"))))
+        yield wordpress1_states['service_relations'][-1].add_unit_state(
+            self.states['unit'])
+        yield wordpress2_states['service_relations'][-1].add_unit_state(
+            self.states['unit'])
 
         # Start and stop lifecycle
         file_path = self.write_start_and_relation_hooks()
         yield self.lifecycle.start()
-        yield self.wait_on_hook("app-relation-changed")
+        yield self.wait_on_hook("app-relation-changed", count=2)
         self.assertTrue(os.path.exists(file_path))
         yield self.lifecycle.stop()
 
@@ -805,10 +825,11 @@ class UnitLifecycleTest(LifecycleTestBase):
                     "wordpress-2", "client-server", "db", "client"))))
 
         yield self.lifecycle.start()
+
         going_workflow = self.lifecycle.get_relation_workflow(
-            self.states["unit_relation"].internal_relation_id)
+            self.states["service_relations"][0].internal_relation_id)
         staying_workflow = self.lifecycle.get_relation_workflow(
-            other_states["unit_relation"].internal_relation_id)
+            other_states["service_relations"][0].internal_relation_id)
 
         # Stop the lifecycle as though the process were being shut down
         yield self.lifecycle.stop(fire_hooks=False, stop_relations=False)
@@ -867,6 +888,38 @@ class UnitLifecycleTest(LifecycleTestBase):
         self.assertEqual("up", (yield staying_workflow.get_state()))
         yield new_lifecycle.stop()
 
+    @inlineCallbacks
+    def test_start_invoker(self):
+        """Verify the invoker is started by the unit relation lifecycle"""
+        # Setup three different wordpress services, wordpress,
+        # wordpress-1, wordpress-2 with one service unit each. Each of
+        # these will be on the relation with ident app:0, app:1, app:2
+        self.write_hook(
+            "start",
+            '#!/bin/sh\n echo "hello world"\n')
+        yield self.add_opposite_service_unit(self.states)
+        for i in range(1, 3):
+            yield self.add_opposite_service_unit(
+                (yield self.add_relation_service_unit_to_another_endpoint(
+                        self.states,
+                        RelationEndpoint(
+                            "wordpress-%d" % i,
+                            "client-server", "db", "client"))))
+
+        finished = self.wait_on_hook("app-relation-changed", 3)
+        yield self.lifecycle.start()
+        yield finished
+
+        # This will verify that config-changed and start both are
+        # cached properly. The remaining hook caching will be for the
+        # relation hook contexts, along some permutation based on how
+        # the hooks are actually scheduled. However, this is tested in
+        # UnitRelationLifecycleTest.test_start_invoker so need not be
+        # covered here.
+        self.assertLogLines(
+            self.hook_log.getvalue(),
+            ["Cached relation hook contexts: ['app:0', 'app:1', 'app:2']"] * 2)
+
 
 class UnitLifecycleUpgradeTest(
         LifecycleTestBase, CharmPublisherTestBase, CharmUpgradeTestBase):
@@ -921,6 +974,7 @@ class UnitLifecycleUpgradeTest(
 
     @inlineCallbacks
     def test_full_run_bad_write(self):
+        self.capture_logging("charm.upgrade")
         yield self.prepare_real_upgrade(0)
         with self.frozen_charm():
             yield self.assertFailure(
@@ -944,16 +998,6 @@ class UnitLifecycleUpgradeTest(
         yield done
         self.assertTrue(self.executor.running)
 
-    @inlineCallbacks
-    def test_full_run_dont_fire_hooks_ignored(self):
-        """Hooks must always be fired if the charm version actually changed
-        in the course of the upgrade"""
-        yield self.prepare_real_upgrade(0)
-        done = self.wait_on_hook("upgrade-charm")
-        yield self.lifecycle.upgrade_charm(fire_hooks=False)
-        yield done
-        self.assertTrue(self.executor.running)
-
 
 class RelationInvokerTest(TestCase):
 
@@ -963,11 +1007,12 @@ class RelationInvokerTest(TestCase):
         self.change_environment(
             PATH=os.environ["PATH"],
             JUJU_UNIT_NAME="service-unit/0")
-        change = RelationChange("clients", "joined", "s/2")
+        change = RelationChange("clients:42", "joined", "s/2")
         unit_hook_path = self.makeDir()
         invoker = RelationInvoker(None, change, "", "", unit_hook_path, None)
         environ = invoker.get_environment()
         self.assertEqual(environ["JUJU_RELATION"], "clients")
+        self.assertEqual(environ["JUJU_RELATION_ID"], "clients:42")
         self.assertEqual(environ["JUJU_REMOTE_UNIT"], "s/2")
         self.assertEqual(environ["CHARM_DIR"],
                          os.path.join(unit_hook_path, "charm"))
@@ -979,6 +1024,7 @@ class UnitRelationLifecycleTest(LifecycleTestBase):
         "#!/bin/bash\n"
         "echo %(change_type)s >> %(file_path)s\n"
         "echo JUJU_RELATION=$JUJU_RELATION >> %(file_path)s\n"
+        "echo JUJU_RELATION_ID=$JUJU_RELATION_ID >> %(file_path)s\n"
         "echo JUJU_REMOTE_UNIT=$JUJU_REMOTE_UNIT >> %(file_path)s")
 
     @inlineCallbacks
@@ -990,7 +1036,7 @@ class UnitRelationLifecycleTest(LifecycleTestBase):
             self.client,
             self.states["unit"].unit_name,
             self.states["unit_relation"],
-            self.states["service_relation"].relation_name,
+            self.states["service_relation"].relation_ident,
             self.unit_directory,
             self.state_directory,
             self.executor)
@@ -1097,13 +1143,16 @@ class UnitRelationLifecycleTest(LifecycleTestBase):
             sequence=["app-relation-joined", "app-relation-changed"])
         self.assertTrue(os.path.exists(file_path))
 
-        contents = open(file_path).read()
+        with open(file_path)as f:
+            contents = f.read()
         self.assertEqual(contents,
                          ("joined\n"
                           "JUJU_RELATION=app\n"
+                          "JUJU_RELATION_ID=app:0\n"
                           "JUJU_REMOTE_UNIT=wordpress/0\n"
                           "changed\n"
                           "JUJU_RELATION=app\n"
+                          "JUJU_RELATION_ID=app:0\n"
                           "JUJU_REMOTE_UNIT=wordpress/0\n"))
 
     @inlineCallbacks
@@ -1208,7 +1257,11 @@ class UnitRelationLifecycleTest(LifecycleTestBase):
         yield hook_complete
         contents = open(file_path).read()
         self.assertEqual(
-            contents, "broken\nJUJU_RELATION=app\nJUJU_REMOTE_UNIT=\n")
+            contents,
+            ("broken\n"
+             "JUJU_RELATION=app\n"
+             "JUJU_RELATION_ID=app:0\n"
+             "JUJU_REMOTE_UNIT=\n"))
 
     @inlineCallbacks
     def test_lock_start_stop(self):
@@ -1267,3 +1320,124 @@ class UnitRelationLifecycleTest(LifecycleTestBase):
 
         # The scheduler should run the hooks it queued up earlier
         yield hooks_complete
+
+    @inlineCallbacks
+    def test_start_invoker(self):
+        """Verify the invoker is started by the unit relation lifecycle"""
+        # Setup 5 different wordpress services, wordpress, wordpress-1
+        # through wordpress-4 with one service unit each. Each of these
+        # will be on the relation app:0 ... app:4
+        yield self.add_opposite_service_unit(self.states)
+        for i in range(1, 5):
+            yield self.add_opposite_service_unit(
+                (yield self.add_relation_service_unit_to_another_endpoint(
+                        self.states,
+                        RelationEndpoint(
+                            "wordpress-%d" % i,
+                            "client-server", "db", "client"))))
+        yield self.lifecycle.start()
+        yield self.wait_on_hook(
+            sequence=["app-relation-joined", "app-relation-changed"])
+
+        # Currently the only action taken by the invoker.start is to
+        # cache the relation hook context; app:0 is excluded because
+        # it is the parent
+        self.assertIn(
+            ("Cached relation hook contexts on 'app:0': "
+             "['app:1', 'app:2', 'app:3', 'app:4']"),
+            self.hook_log.getvalue())
+
+
+class SubordinateUnitLifecycleTest(LifecycleTestBase, CharmPublisherTestBase):
+
+    @inlineCallbacks
+    def setUp(self):
+        yield super(SubordinateUnitLifecycleTest, self).setUp()
+        self.output = self.capture_logging(level=logging.DEBUG)
+        environment = self.config.get_default()
+        self.provider = environment.get_machine_provider()
+        self.storage = self.provider.get_file_storage()
+
+    @inlineCallbacks
+    def get_local_charm(self, charm_name):
+        "override get_local_charm to use copy"
+
+        repo = self.makeDir()
+        os.rmdir(repo)
+        shutil.copytree(unbundled_repository, repo)
+
+        charm_dir = CharmDirectory(
+            os.path.join(repo, "series", charm_name))
+
+        charm, charm_state = yield self.publish_charm(charm_dir.path)
+
+        @self.addCleanup
+        def rm_repo():
+            shutil.rmtree(repo)
+
+        returnValue(charm_state)
+
+    @inlineCallbacks
+    def setup_default_subordinate_relation(self):
+        mysql_ep = RelationEndpoint("mysql", "juju-info", "juju-info",
+                                    "server", "global")
+        logging_ep = RelationEndpoint("logging", "juju-info", "juju-info",
+                                      "client", "container")
+
+        mysql, my_units = yield self.get_service_and_units_by_charm_name(
+            "mysql", 1)
+        self.assertFalse((yield mysql.is_subordinate()))
+
+        log, log_units = yield self.get_service_and_units_by_charm_name(
+            "logging")
+        self.assertTrue((yield log.is_subordinate()))
+
+        # add the relationship so we can create units with  containers
+        relation_state, service_states = (yield
+            self.relation_manager.add_relation_state(
+            mysql_ep, logging_ep))
+
+        mu1 = my_units[0]
+        msm = MachineStateManager(self.client)
+        m1 = yield msm.add_machine_state(series_constraints)
+        yield m1.set_instance_id(str(m1.id))
+        yield mu1.assign_to_machine(m1)
+        yield my_units[0].set_private_address("10.10.10.42")
+
+        self.unit_directory = os.path.join(self.juju_directory,
+            "units",
+            mu1.unit_name.replace("/", "-"))
+
+        os.makedirs(os.path.join(self.unit_directory, "charm", "hooks"))
+        self.state_directory = os.path.join(self.juju_directory, "state")
+        os.makedirs(self.state_directory)
+
+        returnValue(dict(principal=(mu1, mysql),
+                         subordinate=[None, log],
+                         relation_state=relation_state))
+
+    @inlineCallbacks
+    def test_subordinate_lifecycle_start(self):
+        state = yield self.setup_default_subordinate_relation()
+        unit, service = state["principal"]
+
+        lifecycle = UnitLifecycle(
+            self.client, unit, service,
+            self.unit_directory, self.state_directory, self.executor)
+
+        test_deferred = Deferred()
+        results = []
+
+        def test_complete(unit_name, machine_id, charm_dir):
+            results.append([unit_name, machine_id, charm_dir])
+            test_deferred.callback(True)
+
+        mock_unit_deploy = self.mocker.patch(lifecycle)
+        mock_unit_deploy._do_unit_deploy("logging/0", 0,
+                                         MATCH(lambda x: os.path.exists(x)))
+        self.mocker.call(test_complete)
+        self.mocker.replay()
+
+        yield lifecycle.start()
+        yield test_deferred
+        # mocker has already verified the call signature, we're happy
