@@ -355,6 +355,7 @@ class UnitWorkflowState(DiskWorkflowState):
         started = (True, True)
         other = (True, False)
         state = yield self.get_state()
+
         if state == "charm_upgrade_error":
             returnValue(mid_upgrade)
         if state == "started":
@@ -503,6 +504,14 @@ class RelationWorkflowState(DiskWorkflowState):
     def lifecycle(self):
         return self._lifecycle
 
+    def get_relation_info(self):
+        """Return relation info for use in persistence."""
+        rs = {}
+        rs[self._state.internal_relation_id] = dict(
+            relation_name=self.relation_name,
+            relation_scope=self._state.relation_scope)
+        return rs
+
     @inlineCallbacks
     def on_hook_error(self, relation_change, error):
         """Handle relation-change hook errors.
